"""Simple graphML parser
for undirected graphs

supports graphml attributes
   <node id= label= >
   <edge source= target= label= >

I'm not sure if label is an exceptable keyword
so this might not strictly be GraphML compatible.
"""

from xml.sax import handler
from xml.sax import make_parser
import Graph

class GraphMLError(Exception):
    pass

class Node(Graph.Node):
    def __init__(self, id, label=None):
        self.id = id
        Graph.Node.__init__(self, label)

    def __str__(self):
        if self.label:
            return '<node id="%s" label="%s"/>'%(self.id,
                                                 self.label)
        else:
            return '<node id="%s"/>'%self.id        

class Edge(Graph.Edge):
    def __init__(self, id, src, dst, label=None):
        self.id = id
        self.src = src
        self.dst = dst
        Graph.Edge.__init__(self, label)

    def __str__(self):
        # output the node with lowest id first
        if self.src < self.dst:
            src, dst = self.src, self.dst
        else:
            src, dst = self.dst, self.src
            
        if self.label:
            return '<edge source="%s" target="%s" label="%s"/>'%(
                src, dst, self.label)
        else:
            return '<edge source="%s" target="%s"/>'%(
                src, dst)

class GraphML(Graph.Graph):
    def __str__(self):
        """output the graph representation using the node
        and edge symorders to determine the output order of
        the nodes and edges""
        res = ['<graphml>',
               ' <graph edgedefault="undirected">']

        # sort nodes and edges in order of their symorders
        nodes = self.nodes()
        nodes.sort(lambda x,y:cmp(x.symorder, y.symorder))
        # for simplicity we'll reassign the id numbers
        # in the graph
        count = 1
        for n in nodes:
            n.id = "n%s"%count
            res.append("   %s"%n)
            count += 1

        edges = self.edges()
        edges.sort(lambda x,y:cmp(x.symorder, y.symorder))

        for e in edges:
            res.append("   %s"%e)

        res.append(" </graph>")
        res.append("</graphml>")
        return "\n".join(res)
    
class GraphMLParser(handler.ContentHandler):
    def __init__(self):
        self.nodes = []
        self.edges = []
        self.graphs = []
        self._graph = 0

    def startElement(self, name, attrs):
        if name == "graph":
            if self._graph:
                raise GraphMLError("Parser doesn't support subgraphs yet")
            self._graph = 1
            
        if name == "node":
            id = attrs.get("id", len(self.nodes))
            label = attrs.get("label", None)
            self.nodes.append(Node(id, label))

        elif name == "edge":
            id = attrs.get("id", len(self.nodes))
            label = attrs.get("label", None)
            src = attrs["source"]
            dst = attrs["target"]
            self.edges.append(Edge(id, src, dst, label))

    def endElement(self, name):
        if name == "graph":
            self._graph = 0
            g = GraphML()
            ids = {}
            for node in self.nodes:
                ids[node.id] = node
                g.addNode(node)
            for edge in self.edges:
                g.addEdge(ids[edge.src], ids[edge.dst], edge)

            self.graphs.append(g)
            self.nodes = []
            self.edges = []

def parseGraphML(file):
    parser = make_parser()
    dh = GraphMLParser()
    
    parser.setContentHandler(dh)
    parser.parse(file)
    return dh.graphs

if __name__ == "__main__":
    graphs = parseGraphML("undirected2.xml")




    

