//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
// DpObject.hpp
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// Abstract class definition for dynamic solution object.  This class defines
// the interface between the problem and the DpSolution class which is used
// to determine the solution. Inheriting class are required to define the
// following member functions:
//
//   (1) virtual bool operator>(const DpObject & tp) const;
//   (2) virtual bool operator<(const DpObject & tp) const;
//
//   Comparison operators used for sorting the input list to the
//   solution class.  The input list is allowed to be in any order
//   and is sorted at the start of the calculation.
//
//   (3) virtual void setPartner(int index, int stage, const DpObject *dp);
//
//   This function sets the partner class in the solution.  Each object
//   in the solution has a partner that is its "best" choice to pair
//   with to form the optimal solution.
//
//   (4) virtual double evalCost(const DpObject *dp) const;
//   (5) virtual double evalInitialCost(void) const;
//   (6) virtual double evalFinalCost(void) const;
//
//   These functions define the cost associated with paring with a particular
//   partner.  The Initial and Final cost functions are used at the beginning
//   and end of the calculation to account for special conditions at the boundaries.
//   These functions can return zero, if no such conditions exist.
//
//   (7) virtual DpObject *cloneDpObject(void) const;
//
//   The clone function is used to produce copies of the object that are
//   required to form the solution.
//
//   (8) virtual void print(ostream& os) const;
//
//   Print method called by operator<< to print the results.
//
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
#ifndef _DPOBJECT_
#define _DPOBJECT_

#include <string>

using namespace std;

class DpObject
{
   public:
      // constructors/destructor
     DpObject(void);
     DpObject(const DpObject &tp);
     virtual ~DpObject(void);

     // member functions
     string id(void) const;
     int partner(void) const;
     double cost(void) const;
     DpObject *clone(void) const;

     // virtual functions
     virtual void setCost(double cost);
     virtual void setPartner(int index, int stage);

     // Pure virtual functions
     virtual bool operator>(const DpObject & tp) const = 0;
     virtual bool operator<(const DpObject & tp) const = 0;
     virtual void setPartner(int index, int stage, const DpObject *dp) = 0;
     virtual double evalCost(const DpObject *dp) const = 0;
     virtual double evalInitialCost(void) const = 0;
     virtual double evalFinalCost(void) const = 0;

     // friend functions
     friend ostream& operator<<( ostream& os, const DpObject& dp );
     virtual void print(ostream& os) const = 0;
  protected:
      string m_id;          // object identifier
      int    m_partner;     // object partner
      double m_cost;        // object cost
   private:
      virtual DpObject *cloneDpObject(void) const = 0;
      //virtual void print(ostream& os) const = 0;
};

// comparison function used for sorting
inline bool DpObjectPtrLessThan(DpObject *dp1, DpObject *dp2)
{
   return(*dp1 < *dp2);
}
#endif
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//      end       end       end       end       end       end       end
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
