//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
// DpSolution.cpp
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// Class functions for DpSolution class.
// The DpSolution used the abstract class DpObject which contains the interface
// to the problem and determine the solution using dynamic programming.

#include "DpSolution.hpp"
#include<algorithm>

//
// constructors * destructors
//
// Note: this class requires not memory management i.e. does not invoke
// new or delete.  Destructor is virtual since the class is abstract.
//
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Default constructor
//
// Inputs:
//
// const list<DpObject *> &dp - object containing the problem definition
// int                    N   - number of objects desired in solution
//
// ABSTRACT: Initialize solution class for finding the problem solution.
//
DpSolution::DpSolution(const list<DpObject *> &dp, const int &N)
{
   // copy and sort input objects
   Strategy p;
   for(list<DpObject *>::const_iterator iter = dp.begin(); iter != dp.end(); ++iter)
   {
      p.push_back((*iter)->clone());
   }
   sort(p.begin(),p.end(),DpObjectPtrLessThan);

   // create solution stages
   m_stage.resize(N);

   for(int i = 0; i < m_stage.size(); ++i)
   {
     m_stage[i].resize(dp.size());
     // store tracking passes
     int j = 0;
     for(Strategy::iterator iter = p.begin(); (iter != p.end()) && (j < dp.size()); ++iter)
     {
        m_stage[i][j] = (*iter)->clone();
        ++j;
     }
   }

   // delete sorted list
   for(Strategy::iterator piter = p.begin(); piter != p.end(); ++piter)
   {
      delete *piter;
   }


}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Destructor
//
// Inputs:
//
// None
//
// ABSTRACT: Free memory used to computing solution
//
DpSolution::~DpSolution(void)
{
   for(int i = 0; i < m_stage.size(); ++i)
   {
      for(int j = 0; j < m_stage[i].size(); ++j)
      {
         delete m_stage[i][j];
      }
   }
}
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// Class member functions
//
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: solve
//
// Inputs:
//
// None.
//
// Returns:
//
// list<DpObject *> - list of objects in solution
//
// ABSTRACT: Calculate and return the list of object that form the optimal solution
//
list<DpObject *> DpSolution::solve(void)
{
   // check that solution is feasible
   if(m_stage[0].size() <= m_stage.size())
   {
      // no solution, return inputs
      list<DpObject *> ans;
      for(int j = 0; j < m_stage[0].size(); ++j)
      {
         ans.push_front(m_stage[0][j]->clone());
      }
      return(ans);
   }

   // compute stage 1 (initial) strategy/cost
   for(int j = 0; j < m_stage[0].size(); ++j)
   {
      m_stage[0][j]->setCost(m_stage[0][j]->evalInitialCost());
   }

   // compute stage 2 to N-1 strategies/costs
   for(int i = 1; i < m_stage.size(); ++i)
   {
     for(int j = 0; j < m_stage[i].size(); ++j)
     {
        double cmin = m_stage[i][j]->evalCost(m_stage[i-1][0])  + m_stage[i-1][0]->cost();
        int partner = 0;
        for(int k = 0; k < j; ++k)
        {
           double c = m_stage[i][j]->evalCost(m_stage[i-1][k])  + m_stage[i-1][k]->cost();
           if(c < cmin)
           {
              cmin = c;
              partner = k;
           }
        }
        m_stage[i][j]->setCost(cmin);
        m_stage[i][j]->setPartner(partner, i-1, m_stage[i-1][partner]);
     }
   }

   // compute final stage N strategy/cost
   double cmin = m_stage[m_stage.size()-1][0]->cost()+m_stage[m_stage.size()-1][0]->evalFinalCost();
   int    imin = 0;
   for(int k = 0; k < m_stage[m_stage.size()-1].size(); ++k)
   {
     m_stage[m_stage.size()-1][k]->setCost(m_stage[m_stage.size()-1][k]->cost() + m_stage[m_stage.size()-1][k]->evalFinalCost());
     if(m_stage[m_stage.size()-1][k]->cost() < cmin)
     {
        imin = k;
        cmin = m_stage[m_stage.size()-1][k]->cost();
     }
   }

   // extract solution
    list<DpObject *> ans;
    int is = m_stage.size()-1;
    while((imin >= 0) && (is >= 0))
    {
      DpObject *tp = m_stage[0][imin]->clone();
      tp->setCost(m_stage[is][imin]->cost());
      tp->setPartner(m_stage[is][imin]->partner(),imin);
      ans.push_front(tp);
      imin = m_stage[is][imin]->partner();
      --is;
    }

    return(ans);

}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: stages
//
// Inputs:
//
// None.
//
// Returns:
//
// lint - number of stages in the solution strategy
//
// ABSTRACT: Return the number of stages in the solution strategy
//
int DpSolution::stages(void) const
{
   return(m_stage.size());
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: stages
//
// Inputs:
//
// int i - stage index
//
// Returns:
//
// int - number of sub-stages at the indicated input stage of the solution
//       strategy
//
// ABSTRACT: Return the number of sub-stages at the indicated input stage of
//           the solution strategy
//
int DpSolution::stages(int i) const
{
   if(i < 0 || i >= m_stage.size()) return(0);
   else return(m_stage[i].size());
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: stage
//
// Inputs:
//
// int i - stage index
// int j - stage sub-index
//
// Returns:
//
// DpObject * - object at the indicated stage and sub-stage  of the solution
//              strategy
//
// ABSTRACT: Return the object at the indicated stage and sub-stage  of the
//           solution strategy
//
DpObject * DpSolution::stage(int i,int j) const
{
   if(i < 0) i = 0;
   if(i >= m_stage.size()) i = m_stage.size()-1;
   if(j < 0) j = 0;
   if(j >= m_stage[i].size()) j = m_stage[i].size()-1;

   return(m_stage[i][j]);
}
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// Class friend functions
//
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: operator<<
//
// Inputs:
//
// ostream&    os - output stream to write to.
// DpSolution& dp - object to write.
//
// Returns:
//
// ostream& - stream written to.
//
// ABSTRACT: Write object to output stream.
//
ostream& operator<<( ostream& os, const DpSolution& dps )
{
   os << endl;
   os << "Number of Stages: ";
   os.width(3); os.fill(' ');
   os << dps.stages() << endl;
   os << "Number of Choices : ";
   os << dps.stages(0) << endl;
   os << endl;
   for(int i = 0; i < dps.stages(); ++i)
   {
      os << "Stage ";
      os.width(3); os.fill(' ');
      os << i+1 << endl;
      for(int j = 0; j < dps.stages(i); ++j)
      {
         os.width(3); os.fill(' ');
         os << j+1 << " ";
         os.width(3); os.fill(' '); os.precision(8);
         os << *dps.stage(i,j) << endl;
      }
      os << endl;
   }

   return(os);
}
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//      end       end       end       end       end       end       end
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
