//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
// TrackingPass.cpp
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// Class functions for TrackingPass concrete class.
// The TrackingPass class defines the concept of a tracking pass and forms the
// object of the problem that the DpSolution class used to determine the
// solution.
//

#include "TrackingPass.hpp"
#include <cfloat>
const double INFINITE_COST = DBL_MAX;
const double OB2_COST = 0.0030;
const double OB1_COST = 0;


//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// constructors * destructors
//
// Note: this class requires not memory management i.e. does not invoke
// new or delete.  Destructor is virtual since the class is abstract.
//
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Default constructor
//
// Inputs:
//
// None.
//
// ABSTRACT: Initialize data member with values.
//
TrackingPass::TrackingPass(void)
{
   m_id      = "";
   m_partner =-1;
   m_cost    = 0;
   m_refBeg  = 0;
   m_refEnd  = 0;
   m_begin   = 0;
   m_end     = 0;
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Copy constructor
//
// Inputs:
//
// const TrackingPass &tp - tracking pass to copy
//
// ABSTRACT: Create a copy of input tracking pass.
//
TrackingPass::TrackingPass(const TrackingPass &tp)
{
   m_id      = tp.m_id;
   m_partner = tp.m_partner;
   m_cost    = tp.m_cost;
   m_refBeg  = tp.m_refBeg;
   m_refEnd  = tp.m_refEnd;
   m_begin   = tp.m_begin;
   m_end     = tp.m_end;
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Constructor
//
// Inputs:
//
// const Epoch &begin   - beginning of tracking pass
// const Epoch &end     - end of tracking pass
// const Epoch & refBeg - beginning of observation window
// const Epoch &refEnd  - end of observation window
// const string &id     - tracking pass identifier
// double cost          - cost of tracking pass
// int partner          - partner of tracking pass
//
// ABSTRACT: Initialize data member with input values.
//
TrackingPass::TrackingPass(const Epoch &begin,   const Epoch &end,
                           const Epoch & refBeg, const Epoch &refEnd,
                           const string &id, double cost, int partner)
{
   m_id      = id;
   m_partner = partner;
   m_cost    = cost;
   m_refBeg  = refBeg;
   m_refEnd  = refEnd;
   m_begin   = begin;
   m_end     = end;
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Destructor
//
// Inputs:
//
// None.
//
// ABSTRACT: Class destructor.
//
TrackingPass::~TrackingPass(void)
{
}
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// Class member functions
//
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: setPartner
//
// Inputs:
//
// int            index - index of partner
// int            stage - strategy index
// const DpObject *dp   - partner object
//
// Returns:
//
// None.
//
// ABSTRACT: Set the partner associated with the object.
//
void TrackingPass::setPartner(int index, int stage, const DpObject *dp)
{
   const TrackingPass *tp = (TrackingPass *) dp;
   m_partner = index;
   if(tp->m_end > m_end) m_end = tp->m_end;
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: operator>
//
// Inputs:
//
// const DpObject *dp   - comparison object
//
// Returns:
//
// bool - true of input object is less, false otherwise
//
// ABSTRACT: Set the partner associated with the object.
//
bool TrackingPass::operator>(const DpObject &dp) const
{
   const TrackingPass *tp = (TrackingPass *) &dp;
   if(m_begin > tp->m_begin) return(true);
   else return(false);
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: operator<
//
// Inputs:
//
// const DpObject *dp   - comparison object
//
// Returns:
//
// bool - false of input object is less, true otherwise
//
// ABSTRACT: Set the partner associated with the object.
//
bool TrackingPass::operator<(const DpObject &dp) const
{
   const TrackingPass *tp = (TrackingPass *) &dp;
   if(m_begin < tp->m_begin) return(true);
   else return(false);
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: evalCost
//
// Inputs:
//
// const DpObject *dp   - partnering object
//
// Returns:
//
// double - cost of partnering with input object
//
// ABSTRACT: Determine the cost associated with partnering
//           with the input object.
//
double TrackingPass::evalCost(const DpObject *dp) const
{
   const TrackingPass *tp = (TrackingPass *) dp;
   if(*this < *tp || (m_begin == tp->m_begin && m_end == tp->m_end && m_id == tp->m_id) )
      return(INFINITE_COST);
   else if(tp->m_end > m_begin) return(0);
   else
   {
      Duration d = (m_begin - tp->m_end)/(m_refEnd-m_refBeg);
      if(dp->id() == "Observer1") d += OB1_COST;
      if(dp->id() == "Observer2") d += OB2_COST;
      return(d*d*d/12.0);
   }
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: evalInitialCost
//
// Inputs:
//
// None
//
// Returns:
//
// double - initial cost of  object
//
// ABSTRACT: Determine the cost associated with the object
//           at the beginning of the strategy.
//
double TrackingPass::evalInitialCost(void) const
{
   if(m_refBeg > m_begin) return(0);
   else
   {
      Duration d = (m_begin-m_refBeg)/(m_refEnd-m_refBeg);
      if(m_id == "Observer1") d += OB1_COST;
      if(m_id == "Observer2") d += OB2_COST;
      return(d*d*d/3.0);
   }
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: evalFinalCost
//
// Inputs:
//
// None
//
// Returns:
//
// double - final cost of  object
//
// ABSTRACT: Determine the cost associated with the object
//           at the end of the strategy.
//
double TrackingPass::evalFinalCost(void) const
{
   if(m_end > m_refEnd) return(0);
   else
   {
      Duration d = (m_refEnd-m_end)/(m_refEnd-m_refBeg);
      if(m_id == "Observer1") d += OB1_COST;
      if(m_id == "Observer2") d += OB2_COST;
      return(d*d*d/3.0);
   }
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: cloneDpObject
//
// Inputs:
//
// None
//
// Returns:
//
// DpObject * - copy of object
//
// ABSTRACT: Create a copy of the object.
//
DpObject * TrackingPass::cloneDpObject(void) const
{
  TrackingPass *tp = new TrackingPass(*this);
  return tp;
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: setEndTime
//
// Inputs:
//
// Epoch &end - end time for tracking pass
//
// Returns:
//
// None
//
// ABSTRACT: Set the end time of the tracking pass
//
void TrackingPass::setEndTime(Epoch &end)
{
   m_end = end;
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: setBeginTime
//
// Inputs:
//
// Epoch &begin - beginning time for tracking pass
//
// Returns:
//
// None
//
// ABSTRACT: Set the beginning time of the tracking pass
//
void TrackingPass::setBeginTime(Epoch &begin)
{
   m_begin = begin;
}
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: print
//
// Inputs:
//
// ostream& os - output stream to print to.
//
// Returns:
//
// None
//
// ABSTRACT: Write object to output stream.
//
void TrackingPass::print(ostream& os) const
{
   os << m_id << ": \t[" << m_begin << ", " << m_end << "] \t" << m_partner+1
      << " \t" << m_cost;
}
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// Class friend functions
//
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
//
// Function: operator<<
//
// Inputs:
//
// ostream&            os - output stream to write to.
// const TrackingPass& tp - object to write.
//
// Returns:
//
// ostream& - stream written to.
//
// ABSTRACT: Write object to output stream.
//
ostream& operator<<( ostream& os, const TrackingPass& tp )
{
   tp.print(os);
   return(os);
}
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//      end       end       end       end       end       end       end
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].

