//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
// TrackingPass.hpp
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//
// Concrete class definition for dynamic solution object.  This class
// encapsulates the concept of a tracking pass.  That is a time interval
// which conditions are right of observe another object such as a
// planet or spacecraft. This class inherits from DpObject and defines
// the following functions:
//
//   (1) virtual bool operator>(const DpObject & tp) const;
//   (2) virtual bool operator<(const DpObject & tp) const;
//
//   Comparison operators used for sorting the input list to the
//   solution class.  The input list is allowed to be in any order
//   and is sorted at the start of the calculation.
//
//   (3) virtual void setPartner(int index, int stage, const DpObject *dp);
//
//   This function sets the partner class in the solution.  Each object
//   in the solution has a partner that is its "best" choice to pair
//   with to form the optimal solution.
//
//   (4) virtual double evalCost(const DpObject *dp) const;
//   (5) virtual double evalInitialCost(void) const;
//   (6) virtual double evalFinalCost(void) const;
//
//   These functions define the cost associated with paring with a particular
//   partner.  The Initial and Final cost functions are used at the beginning
//   and end of the calculation to account for special conditions at the boundaries.
//   These functions can return zero, if no such conditions exist.
//
//   (7) virtual DpObject *cloneDpObject(void) const;
//
//   The clone function is used to produce copies of the object that are
//   required to form the solution.
//
//   (8) virtual void print(ostream& os) const;
//
//   Print method called by operator<< to print the results.
//
//..............................................................................
//       1         2         3         4         5         6         7         8
//345678901234567890123456789012345678901234567890123456789012345678901234567890
//..............................................................................
#ifndef _TRACKING_PASS_
#define _TRACKING_PASS_

#include <string>
#include "DpObject.hpp"

typedef double Epoch;
typedef double Duration;

class TrackingPass : public DpObject
{
   public:
      // constructors/destructor
     TrackingPass(void);
     TrackingPass(const TrackingPass &tp);
     TrackingPass(const Epoch &begin,    const Epoch &end,
                  const Epoch &refBegin, const Epoch &refEnd,
                  const string &id, double cost = 0, int partner = -1 );
     virtual ~TrackingPass(void);

     // functions to set beginning and end of tracking pass interval
     void setEndTime(Epoch &end);
     void setBeginTime(Epoch &begin);

     // virtual functions from base class
     virtual void setPartner(int index, int stage, const DpObject *dp);
     virtual double evalCost(const DpObject *dp) const;

     virtual double evalInitialCost(void) const;
     virtual double evalFinalCost(void) const;

     virtual bool operator>(const DpObject &dp) const;
     virtual bool operator<(const DpObject &dp) const;
     friend ostream& operator<<( ostream& os, const TrackingPass& tp );
     virtual void print(ostream& os) const;
   private:
      virtual DpObject * cloneDpObject(void) const;
      //virtual void print(ostream& os) const;
      Epoch m_refBeg; // start of observation window
      Epoch m_refEnd; // end of observation window
      Epoch m_begin;  // beginning of tracking pass
      Epoch m_end;    // end of tracking pass
};
#endif
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].
//      end       end       end       end       end       end       end
//......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......]*[.......].